/*
 * $Id: genpkglist.cc,v 1.7 2003/01/30 17:18:21 niemeyer Exp $
 */
#include <alloca.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <assert.h>

#include <map>
#include <iostream>

#include <apt-pkg/error.h>
#include <apt-pkg/tagfile.h>
#include <apt-pkg/configuration.h>
#include <config.h>

#include "rpmhandler.h"
#include "cached_md5.h"
#include "genutil.h"

#include <rpmts.h>
#include <rpmrc.h>
#include <pkgio.h>
#include <rpmlog.h>

#define CRPMTAG_TIMESTAMP   1012345

int tags[] =  {
       RPMTAG_NAME, 
       RPMTAG_EPOCH,
       RPMTAG_VERSION,
       RPMTAG_RELEASE,
       RPMTAG_GROUP,
       RPMTAG_ARCH,
       RPMTAG_PACKAGER,
       RPMTAG_SOURCERPM,
       RPMTAG_SIZE,
       RPMTAG_VENDOR,
       RPMTAG_OS,
       
       RPMTAG_DESCRIPTION, 
       RPMTAG_SUMMARY, 
       /*RPMTAG_HEADERI18NTABLE*/ HEADER_I18NTABLE,
       
       RPMTAG_REQUIREFLAGS, 
       RPMTAG_REQUIRENAME,
       RPMTAG_REQUIREVERSION,
       
       RPMTAG_CONFLICTFLAGS,
       RPMTAG_CONFLICTNAME,
       RPMTAG_CONFLICTVERSION,
       
       RPMTAG_PROVIDENAME,
       RPMTAG_PROVIDEFLAGS,
       RPMTAG_PROVIDEVERSION,
       
       RPMTAG_OBSOLETENAME,
       RPMTAG_OBSOLETEFLAGS,
       RPMTAG_OBSOLETEVERSION
};
int numTags = sizeof(tags) / sizeof(int);



typedef struct {
   string importance;
   string date;
   string summary;
   string url;
} UpdateInfo;


static
int usefulFile(const char *d, const char *b)
{
   // PATH-like directories
   if (endswith(d, "/bin/") || endswith(d, "/sbin/"))
      return 1;

   // shared libraries
   if (strncmp(b, "lib", 3) == 0 && strstr(b + 3, ".so"))
      return 1;

   return 0;
}


static void copyStrippedFileList(Header header, Header newHeader)
{
   HE_t he1 = (HE_t)memset(alloca(sizeof(*he1)), 0, sizeof(*he1));
   HE_t he2 = (HE_t)memset(alloca(sizeof(*he2)), 0, sizeof(*he2));
   HE_t he3 = (HE_t)memset(alloca(sizeof(*he3)), 0, sizeof(*he3));

   rpmTagCount i, i1, i2;
   rpmTagType type1, type2, type3;
   rpmTagCount count1, count2, count3;
   char **dirnames = NULL, **basenames = NULL;
   uint32_t *dirindexes = NULL;
   rpmTagData dirnameval, basenameval, dirindexval;
   char **dnames, **bnames;
   uint32_t *dindexes;
   int res1, res2, res3;
   
   he1->tag = RPMTAG_DIRNAMES;
   res1 = headerGet(header, he1, RPMTAG_DIRNAMES);
   if (res1) {
       type1 = he1->t;
       dirnameval.argv = he1->p.argv;
       count1 = he1->c;
   }
   he2->tag = RPMTAG_BASENAMES;
   res2 = headerGet(header, he2, RPMTAG_BASENAMES);
   if (res2) {
       type2 = he2->t;
       basenameval.argv = he2->p.argv;
       count2 = he2->c;
   }
   he3->tag = RPMTAG_DIRINDEXES;
   res3 = headerGet(header, he3, RPMTAG_DIRINDEXES);
   if (res3) {
       type3 = he3->t;
       dirindexval.ui32p = he3->p.ui32p;
       count3 = he3->c;
   }
   
   if (res1 != 1 || res2 != 1 || res3 != 1) {
      _free(dirnames);
      _free(basenames);
      return;
   }

   dnames = dirnames;
   bnames = basenames;
   dindexes = (uint32_t*)malloc(sizeof(uint32_t)*count3);
   
   i1 = 0;
   i2 = 0;
   for (i = 0; i < count2 ; i++) 
   {
      int ok = usefulFile(dirnames[dirindexes[i]], basenames[i]);
      
      if (!ok) {
	 int k = i;
	 while (dirindexes[i] == dirindexes[k] && i < count2)
	     i++;
	 i--;
	 continue;
      }
      
      
      if (ok)
      {
	 rpmTagCount j;
	 
	 bnames[i1] = basenames[i];
	 for (j = 0; j < i2; j++)
	 {
	    if (dnames[j] == dirnames[dirindexes[i]])
	    {
	       dindexes[i1] = j;
	       break;
	    }
	 }
	 if (j == i2) 
	 {
	    dnames[i2] = dirnames[dirindexes[i]];
	    dindexes[i1] = i2;
	    i2++;
	 }
	 assert(i2 <= count1);
	 i1++;
      } 
   }
   
   if (i1 == 0) {
      _free(dirnames);
      _free(basenames);
      _free(dindexes);
      return;
   }
   

   he1->tag = RPMTAG_DIRNAMES;
   he1->t = type1;
   he1->p.argv = (const char**)dnames;
   he1->c = i2;
   headerPut(newHeader, he1, 0);

   he2->tag = RPMTAG_BASENAMES;
   he2->t = type2;
   he2->p.argv = (const char**)bnames;
   he2->c = i1;
   headerPut(newHeader, he2, 0);

   he3->tag = RPMTAG_DIRINDEXES;
   he3->t = type3;
   he3->p.ui32p = dindexes;
   he3->c = i1;
   headerPut(newHeader, he3, 0);

   _free(dirnames);
   _free(basenames);
   _free(dindexes);
}





bool loadUpdateInfo(char *path, map<string,UpdateInfo> &map)
{
   FileFd F(path, FileFd::ReadOnly);
   if (_error->PendingError()) 
   {
      return false;
   }
   
   pkgTagFile Tags(&F);
   pkgTagSection Section;
   
   while (Tags.Step(Section)) 
   {
      string file = Section.FindS("File");
      UpdateInfo info;

      info.importance = Section.FindS("Importance");
      info.date = Section.FindS("Date");
      info.summary = Section.FindS("Summary");
      info.url = Section.FindS("URL");

      map[file] = info;
   }
   return true;
}

bool copyFields(Header h, Header newHeader,
		FILE *idxfile, const char *directory, char *filename,
		unsigned filesize, map<string,UpdateInfo> &updateInfo,
		bool fullFileList)
{
   int i;
   uint32_t size[1];
   HE_t he = (HE_t)memset(alloca(sizeof(*he)), 0, sizeof(*he));

   size[0] = filesize;
   
   // the std tags
   for (i = 0; i < numTags; i++) {
       rpmTagType type;
       rpmTagCount count;
       rpmTagData data;
       int res;

       // Copy raw entry, so that internationalized strings
       // will get copied correctly.
       he->tag = (rpmTag)tags[i];
       he->t = type;
       he->p.ptr = data.ptr;
       he->c = count;
       res = headerGet(h, he, tags[i]);
       if (res) {
	   type = he->t;
	   data.ptr = he->p.ptr;
	   count = he->c;
       }
       else
	   continue;

       he->tag = (rpmTag)tags[i];
       he->t = type;
       he->p.ptr = data.ptr;
       he->c = count;
       headerPut(newHeader, he, 0);
   }
 
   if (fullFileList) {
       HE_t he1 = (HE_t)memset(alloca(sizeof(*he1)), 0, sizeof(*he1));
       HE_t he2 = (HE_t)memset(alloca(sizeof(*he2)), 0, sizeof(*he2));
       HE_t he3 = (HE_t)memset(alloca(sizeof(*he3)), 0, sizeof(*he3));

      rpmTagType type1, type2, type3;
      rpmTagCount count1, count2, count3;
      char **dnames, **bnames, **dindexes;
      rpmTagData dnameval, bnameval, dindexval;
      int res;
   

      he1->tag = RPMTAG_DIRNAMES;
      res = headerGet(h, he1, RPMTAG_DIRNAMES);
      if (res) {
	  type1 = he1->t;
	  dnames = (char **)he1->p.argv;
	  count1 = he1->c;
      }
      he2->tag = RPMTAG_BASENAMES;
      res = headerGet(h, he2, 0);
      if (res) {
	  type2 = he2->t;
	  bnames = (char **)he2->p.argv;
	  count2 = he2->c;
      }
      he3->tag = RPMTAG_DIRINDEXES;
      res = headerGet(h, he3, 0);
      if (res) {
	  type3 = he3->t;
	  dindexes = (char **)he3->p.argv;
	  count3 = he3->c;
      }


      if (res == 1) {

	  he1->tag = RPMTAG_DIRNAMES;
	  he1->t = type1;
	  he1->p.argv = (const char**)dnames;
	  he1->c = count1;
	  headerPut(newHeader, he1, 0);

	  he2->tag = RPMTAG_BASENAMES;
	  he2->t = type2;
	  he2->p.argv = (const char**)bnames;
	  he2->c = count2;
	  headerPut(newHeader, he2, 0);

	  he3->tag = RPMTAG_DIRINDEXES;
	  he3->t = type3;
	  he3->p.argv = (const char**)dindexes;
	  he3->c = count3;
	  headerPut(newHeader, he3, 0);
      }
   } else {
       copyStrippedFileList(h, newHeader);
   }

   // update index of srpms
   if (idxfile) {
       rpmTagType type;
       rpmTagCount count;
       rpmTagData srpmval, nameval;
       char *srpm, *name;
       int res;
       HE_t he = (HE_t)memset(alloca(sizeof(*he)), 0, sizeof(*he));
       
       he->tag = RPMTAG_NAME;
       res = headerGet(h, he, 0);
       if (res) {
	   type = he->t;
	   nameval.str = he->p.str;
	   count = he->c;
       }
       name = (char*)nameval.str;

       he->tag = RPMTAG_SOURCERPM;
       res = headerGet(h, he, 0);
       if (res) {
	   type = he->t;
	   srpmval.str = he->p.str;
	   count = he->c;
       }
       srpm = (char*)srpmval.str;
      
      if (res == 1) {
	 fprintf(idxfile, "%s %s\n", srpm, name);
      }
   }
   // our additional tags

	he->tag = CRPMTAG_DIRECTORY;
	he->t = RPM_STRING_TYPE;
	he->p.str = (const char*)directory;
	he->c = 1;
	headerPut(newHeader, he, 0);

	he->tag = CRPMTAG_FILENAME;
	he->t = RPM_STRING_TYPE;
	he->p.str = (const char*)filename;
	he->c = 1;
	headerPut(newHeader, he, 0);

	he->tag = CRPMTAG_FILESIZE;
	he->t = RPM_UINT32_TYPE;
	he->p.ui32p = size;
	he->c = 1;
	headerPut(newHeader, he, 0);

      
   // update description tags
   if (updateInfo.find(string(filename)) != updateInfo.end()) {
      const char *tmp;
      string name = string(filename);

      tmp = updateInfo[name].summary.c_str();
      he->tag = CRPMTAG_UPDATE_SUMMARY;
      he->t = RPM_STRING_TYPE;
      he->p.str = tmp;
      he->c = 1;
      headerPut(newHeader, he, 0);

      tmp = updateInfo[name].url.c_str();
      he->tag = CRPMTAG_UPDATE_URL;
      he->t = RPM_STRING_TYPE;
      he->p.str = tmp;
      he->c = 1;
      headerPut(newHeader, he, 0);


      tmp = updateInfo[name].date.c_str();
      he->tag = CRPMTAG_UPDATE_DATE;
      he->t = RPM_STRING_TYPE;
      he->p.str = tmp;
      he->c = 1;
      headerPut(newHeader, he, 0);

      tmp = updateInfo[name].importance.c_str();
      he->tag = CRPMTAG_UPDATE_IMPORTANCE;
      he->t = RPM_STRING_TYPE;
      he->p.str = tmp;
      he->c = 1;
      headerPut(newHeader, he, 0);
   }
   
   return true;
}


void usage()
{
   cerr << "genpkglist " << VERSION << endl;
   cerr << "usage: genpkglist [<options>] <dir> <suffix>" << endl;
   cerr << "options:" << endl;
   cerr << " --index <file>  file to write srpm index data to" << endl;
   cerr << " --info <file>   file to read update info from" << endl;
   cerr << " --meta <suffix> create package file list with given suffix" << endl;
   cerr << " --bloat         do not strip the package file list. Needed for some" << endl;
   cerr << "                 distributions that use non-automatically generated" << endl;
   cerr << "                 file dependencies" << endl;
   cerr << " --append        append to the package file list, don't overwrite" << endl;
   cerr << " --progress      show a progress bar" << endl;
   cerr << " --cachedir=DIR  use a custom directory for package md5sum cache"<<endl;
}


int main(int argc, char ** argv) 
{
   string rpmsdir;
   string pkglist_path;
   FD_t outfd, fd;
   struct dirent **dirEntries;
   int entry_no, entry_cur;
   map<string,UpdateInfo> updateInfo;
   CachedMD5 *md5cache;
   char *op_dir;
   char *op_suf;
   char *op_index = NULL;
   char *op_update = NULL;
   FILE *idxfile;
   int i;
   bool fullFileList = false;
   bool progressBar = false;
   const char *pkgListSuffix = NULL;
   bool pkgListAppend = false;
   
   putenv((char *)"LC_ALL="); // Is this necessary yet (after i18n was supported)?
   for (i = 1; i < argc; i++) {
      if (strcmp(argv[i], "--index") == 0) {
	 i++;
	 if (i < argc) {
	    op_index = argv[i];
	 } else {
	    cout << "genpkglist: filename missing for option --index"<<endl;
	    exit(1);
	 }
      } else if (strcmp(argv[i], "--info") == 0) {
	 i++;
	 if (i < argc) {
	    op_update = argv[i];
	 } else {
	    cout << "genpkglist: filename missing for option --info"<<endl;
	    exit(1);
	 }
      } else if (strcmp(argv[i], "--bloat") == 0) {
	 fullFileList = true;
      } else if (strcmp(argv[i], "--progress") == 0) {
	 progressBar = true;
      } else if (strcmp(argv[i], "--append") == 0) {
	 pkgListAppend = true;
      } else if (strcmp(argv[i], "--meta") == 0) {
	 i++;
	 if (i < argc) {
	    pkgListSuffix = argv[i];
	 } else {
	    cout << "genpkglist: argument missing for option --meta"<<endl;
	    exit(1);
	 }
      } else if (strcmp(argv[i], "--cachedir") == 0) {
	 i++;
	 if (i < argc) {
            _config->Set("Dir::Cache", argv[i]);
	 } else {
            cout << "genpkglist: argument missing for option --cachedir"<<endl;
	    exit(1);
	 }
      } else {
	 break;
      }
   }
   if (argc - i > 0)
       op_dir = argv[i++];
   else {
      usage();
      exit(1);
   }
   if (argc - i > 0)
       op_suf = argv[i++];
   else {
      usage();
      exit(1);
   }
   if (argc != i) {
      usage();
   }
   
   if (op_update) {
      if (!loadUpdateInfo(op_update, updateInfo)) {
	 cerr << "genpkglist: error reading update info from file " << op_update << endl;
	 _error->DumpErrors();
	 exit(1);
      }
   }
   if (op_index) {
      idxfile = fopen(op_index, "w+");
      if (!idxfile) {
	 cerr << "genpkglist: could not open " << op_index << " for writing";
	 perror("");
	 exit(1);
      }
   } else {
      idxfile = NULL;
   }
   
   {
      char cwd[PATH_MAX];
      
      if (getcwd(cwd, PATH_MAX) == 0)
      {
         cerr << argv[0] << ": " << strerror(errno) << endl;
         exit(1);
      }
      if (*op_dir != '/') {
	 rpmsdir = string(cwd) + "/" + string(op_dir);
      } else {
	 rpmsdir = string(op_dir);
      }
   }
   pkglist_path = string(rpmsdir);
   rpmsdir = rpmsdir + "/RPMS." + string(op_suf);

   string dirtag = "RPMS." + string(op_suf);

   entry_no = scandir(rpmsdir.c_str(), &dirEntries, selectRPMs, alphasort);
   if (entry_no < 0) {
      cerr << "genpkglist: error opening directory " << rpmsdir << ": "
	  << strerror(errno) << endl;
      return 1;
   }
   
   if (chdir(rpmsdir.c_str()) != 0)
   {
      cerr << argv[0] << ": " << strerror(errno) << endl;
      return 1;
   }
   
   if (pkgListSuffix != NULL)
	   pkglist_path = pkglist_path + "/base/pkglist." + pkgListSuffix;
   else
	   pkglist_path = pkglist_path + "/base/pkglist." + op_suf;
   
   
   if (pkgListAppend == true && FileExists(pkglist_path)) {
      outfd = Fopen(pkglist_path.c_str(), "a");
   } else {
      unlink(pkglist_path.c_str());
      outfd = Fopen(pkglist_path.c_str(), "w+");
   }
   if (!outfd) {
      cerr << "genpkglist: error creating file " << pkglist_path << ": "
	  << strerror(errno) << endl;
      return 1;
   }

   md5cache = new CachedMD5(string(op_dir) + string(op_suf), "genpkglist");

   rpmReadConfigFiles(NULL, NULL);
   rpmts ts = rpmtsCreate();
   rpmtsSetVSFlags(ts, (pgpVSFlags_e)-1);

   for (entry_cur = 0; entry_cur < entry_no; entry_cur++) {
      struct stat sb;

      if (progressBar) {
         if (entry_cur)
            printf("\b\b\b\b\b\b\b\b\b\b");
         printf(" %04i/%04i", entry_cur + 1, entry_no);
         fflush(stdout);
      }

      if (stat(dirEntries[entry_cur]->d_name, &sb) < 0) {
	    cerr << "\nWarning: " << strerror(errno) << ": " << 
		    dirEntries[entry_cur]->d_name << endl;
	    continue;
      }

      {
	 Header h;
	 int rc;
	 
	 fd = Fopen(dirEntries[entry_cur]->d_name, "r");

	 if (!fd) {
	     cerr << "\nWarning: " << strerror(errno) << ": " << 
		 dirEntries[entry_cur]->d_name << endl;
	     continue;
	 }

	 rc = rpmReadPackageFile(ts, fd, dirEntries[entry_cur]->d_name, &h);
	 if (rc == RPMRC_OK || rc == RPMRC_NOTTRUSTED || rc == RPMRC_NOKEY) {
	     HE_t he = (HE_t)memset(alloca(sizeof(*he)), 0, sizeof(*he));
	     Header newHeader;
	     char md5[34];

	     newHeader = headerNew();

	     copyFields(h, newHeader, idxfile, dirtag.c_str(),
		     dirEntries[entry_cur]->d_name,
		     sb.st_size, updateInfo, fullFileList);

	     md5cache->MD5ForFile(string(dirEntries[entry_cur]->d_name), 
		     sb.st_mtime, md5);
	     he->tag = CRPMTAG_MD5;
	     he->t = RPM_STRING_TYPE;
	     he->p.str = md5;
	     he->c = 1;
	     headerPut(newHeader, he, 0);


	     const char * msg = NULL;
	     const char item[] = "Header";
	     rpmRC rc = rpmpkgWrite(item, outfd, newHeader, &msg);
	     if (rc != RPMRC_OK) {
		 rpmlog(RPMLOG_ERR, "%s: %s\n", item, msg);
		 rc = RPMRC_FAIL;
	     }
	     msg = (const char*)_free(msg);

	     headerFree(newHeader);
	     headerFree(h);
	 } else {
	     cerr << "\nWarning: Skipping malformed RPM: " << 
		 dirEntries[entry_cur]->d_name << endl;
	 }
	 Fclose(fd);
      }
   }

   Fclose(outfd);

   ts = rpmtsFree(ts);

   delete md5cache;

   return 0;
}
